package org.eclipse.jdt.internal.compiler.flow;

package .tmp;

import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.LabelFlowContext;
import org.eclipse.jdt.internal.compiler.flow.SwitchFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;

public class LoopingFlowContext extends SwitchFlowContext {
  public BranchLabel continueLabel;
  
  public UnconditionalFlowInfo initsOnContinue = FlowInfo.DEAD_END;
  
  private UnconditionalFlowInfo upstreamNullFlowInfo;
  
  private org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext[] innerFlowContexts = null;
  
  private UnconditionalFlowInfo[] innerFlowInfos = null;
  
  private int innerFlowContextsCount = 0;
  
  private LabelFlowContext[] breakTargetContexts = null;
  
  private int breakTargetsCount = 0;
  
  Reference[] finalAssignments;
  
  VariableBinding[] finalVariables;
  
  int assignCount = 0;
  
  LocalVariableBinding[] nullLocals;
  
  ASTNode[] nullReferences;
  
  int[] nullCheckTypes;
  
  UnconditionalFlowInfo[] nullInfos;
  
  int nullCount;
  
  private ArrayList escapingExceptionCatchSites = null;
  
  Scope associatedScope;
  
  public LoopingFlowContext(FlowContext parent, FlowInfo upstreamNullFlowInfo, ASTNode associatedNode, BranchLabel breakLabel, BranchLabel continueLabel, Scope associatedScope, boolean isPreTest) {
    super(parent, associatedNode, breakLabel, isPreTest);
    this.tagBits |= 0x2;
    this.continueLabel = continueLabel;
    this.associatedScope = associatedScope;
    this.upstreamNullFlowInfo = upstreamNullFlowInfo.unconditionalCopy();
  }
  
  public void complainOnDeferredFinalChecks(BlockScope scope, FlowInfo flowInfo) {
    for (int i = 0; i < this.assignCount; i++) {
      VariableBinding variable = this.finalVariables[i];
      if (variable != null) {
        boolean complained = false;
        if (variable instanceof FieldBinding) {
          if (flowInfo.isPotentiallyAssigned((FieldBinding)variable)) {
            complained = true;
            scope.problemReporter().duplicateInitializationOfBlankFinalField(
                (FieldBinding)variable, 
                this.finalAssignments[i]);
          } 
        } else if (flowInfo.isPotentiallyAssigned((LocalVariableBinding)variable)) {
          variable.tagBits &= 0xFFFFFFFFFFFFF7FFL;
          if (variable.isFinal()) {
            complained = true;
            scope.problemReporter().duplicateInitializationOfFinalLocal(
                (LocalVariableBinding)variable, 
                (ASTNode)this.finalAssignments[i]);
          } 
        } 
        if (complained) {
          FlowContext context = getLocalParent();
          while (context != null) {
            context.removeFinalAssignmentIfAny(this.finalAssignments[i]);
            context = context.getLocalParent();
          } 
        } 
      } 
    } 
  }
  
  public void complainOnDeferredNullChecks(BlockScope scope, FlowInfo callerFlowInfo) {
    complainOnDeferredNullChecks(scope, callerFlowInfo, true);
  }
  
  public void complainOnDeferredNullChecks(BlockScope scope, FlowInfo callerFlowInfo, boolean updateInitsOnBreak) {
    for (int i = 0; i < this.innerFlowContextsCount; i++)
      this.upstreamNullFlowInfo
        .addPotentialNullInfoFrom(
          (this.innerFlowContexts[i]).upstreamNullFlowInfo)
        .addPotentialNullInfoFrom(this.innerFlowInfos[i]); 
    this.innerFlowContextsCount = 0;
    FlowInfo upstreamCopy = this.upstreamNullFlowInfo.copy();
    UnconditionalFlowInfo incomingInfo = this.upstreamNullFlowInfo
      .addPotentialNullInfoFrom(callerFlowInfo.unconditionalInitsWithoutSideEffect());
    if ((this.tagBits & 0x1) != 0) {
      for (int j = 0; j < this.nullCount; j++) {
        Expression expression;
        int nullStatus;
        FakedTrackingVariable trackingVar;
        LocalVariableBinding local = this.nullLocals[j];
        ASTNode location = this.nullReferences[j];
        FlowInfo flowInfo = (this.nullInfos[j] != null) ? 
          incomingInfo.copy().addNullInfoFrom((FlowInfo)this.nullInfos[j]) : 
          (FlowInfo)incomingInfo;
        switch (this.nullCheckTypes[j] & 0xFFFF0FFF) {
          case 258:
          case 514:
            if (flowInfo.isDefinitelyNonNull(local)) {
              this.nullReferences[j] = null;
              if ((this.nullCheckTypes[j] & 0xFFFF0FFF) == 514) {
                if ((this.nullCheckTypes[j] & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNonNull(local, location); 
                break;
              } 
              scope.problemReporter().localVariableNonNullComparedToNull(local, location);
              break;
            } 
          case 256:
          case 512:
            if (flowInfo.isDefinitelyNonNull(local)) {
              this.nullReferences[j] = null;
              if ((this.nullCheckTypes[j] & 0xFFFF0FFF) == 512) {
                if ((this.nullCheckTypes[j] & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNonNull(local, location); 
                break;
              } 
              scope.problemReporter().localVariableNonNullComparedToNull(local, location);
              break;
            } 
            if (flowInfo.isDefinitelyNull(local)) {
              this.nullReferences[j] = null;
              if ((this.nullCheckTypes[j] & 0xFFFF0FFF) == 256) {
                if ((this.nullCheckTypes[j] & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNull(local, location); 
                break;
              } 
              scope.problemReporter().localVariableNullComparedToNonNull(local, location);
              break;
            } 
          case 257:
          case 513:
          case 769:
          case 1025:
            expression = (Expression)location;
            if (flowInfo.isDefinitelyNull(local)) {
              this.nullReferences[j] = null;
              switch (this.nullCheckTypes[j] & 0xFFFF0F00) {
                case 256:
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)expression);
                    break;
                  } 
                  if ((this.nullCheckTypes[j] & 0x1000) == 0)
                    scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)expression); 
                  break;
                case 512:
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)expression);
                    break;
                  } 
                  scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)expression);
                  break;
                case 768:
                  scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)expression);
                  break;
                case 1024:
                  scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)expression);
                  break;
              } 
            } else if (flowInfo.isPotentiallyNull(local)) {
              switch (this.nullCheckTypes[j] & 0xFFFF0F00) {
                case 256:
                  this.nullReferences[j] = null;
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0)
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)expression); 
                  break;
                case 512:
                  this.nullReferences[j] = null;
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0)
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)expression); 
                  break;
              } 
            } 
          case 3:
            if (flowInfo.isDefinitelyNull(local)) {
              this.nullReferences[j] = null;
              scope.problemReporter().localVariableNullReference(local, location);
              break;
            } 
          case 128:
            nullStatus = flowInfo.nullStatus(local);
            if (nullStatus != 4)
              this.parent.recordNullityMismatch(scope, (Expression)location, this.providedExpectedTypes[j][0], this.providedExpectedTypes[j][1], flowInfo, nullStatus, null); 
            break;
          case 2048:
            trackingVar = local.closeTracker;
            if (trackingVar != null) {
              if (trackingVar.hasDefinitelyNoResource(flowInfo))
                break; 
              if (trackingVar.isClosedInFinallyOfEnclosing(scope))
                break; 
              if (this.parent.recordExitAgainstResource(scope, flowInfo, trackingVar, location)) {
                this.nullReferences[j] = null;
                break;
              } 
            } 
          case 16:
            checkUnboxing((Scope)scope, (Expression)location, flowInfo);
            break;
          default:
            if (this.nullCheckTypes[j] != 3 || !upstreamCopy.isDefinitelyNonNull(local))
              this.parent.recordUsingNullReference((Scope)scope, local, location, 
                  this.nullCheckTypes[j], flowInfo); 
            break;
        } 
      } 
    } else {
      for (int j = 0; j < this.nullCount; j++) {
        Expression expression;
        int nullStatus;
        ASTNode location = this.nullReferences[j];
        LocalVariableBinding local = this.nullLocals[j];
        FlowInfo flowInfo = (this.nullInfos[j] != null) ? 
          incomingInfo.copy().addNullInfoFrom((FlowInfo)this.nullInfos[j]) : 
          (FlowInfo)incomingInfo;
        switch (this.nullCheckTypes[j] & 0xFFFF0FFF) {
          case 256:
          case 512:
            if (flowInfo.isDefinitelyNonNull(local)) {
              this.nullReferences[j] = null;
              if ((this.nullCheckTypes[j] & 0xFFFF0FFF) == 512) {
                if ((this.nullCheckTypes[j] & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNonNull(local, location); 
                break;
              } 
              scope.problemReporter().localVariableNonNullComparedToNull(local, location);
              break;
            } 
          case 257:
          case 513:
          case 769:
          case 1025:
            expression = (Expression)location;
            if (flowInfo.isDefinitelyNull(local)) {
              this.nullReferences[j] = null;
              switch (this.nullCheckTypes[j] & 0xFFFF0F00) {
                case 256:
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)expression);
                    break;
                  } 
                  if ((this.nullCheckTypes[j] & 0x1000) == 0)
                    scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)expression); 
                  break;
                case 512:
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)expression);
                    break;
                  } 
                  scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)expression);
                  break;
                case 768:
                  scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)expression);
                  break;
                case 1024:
                  scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)expression);
                  break;
              } 
              break;
            } 
            if (flowInfo.isPotentiallyNull(local))
              switch (this.nullCheckTypes[j] & 0xFFFF0F00) {
                case 256:
                  this.nullReferences[j] = null;
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0)
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)expression); 
                  break;
                case 512:
                  this.nullReferences[j] = null;
                  if ((this.nullCheckTypes[j] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0)
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)expression); 
                  break;
              }  
            break;
          case 3:
            if (flowInfo.isDefinitelyNull(local)) {
              this.nullReferences[j] = null;
              scope.problemReporter().localVariableNullReference(local, location);
              break;
            } 
            if (flowInfo.isPotentiallyNull(local)) {
              this.nullReferences[j] = null;
              scope.problemReporter().localVariablePotentialNullReference(local, location);
            } 
            break;
          case 128:
            nullStatus = flowInfo.nullStatus(local);
            if (nullStatus != 4) {
              char[][] annotationName = scope.environment().getNonNullAnnotationName();
              scope.problemReporter().nullityMismatch((Expression)location, this.providedExpectedTypes[j][0], this.providedExpectedTypes[j][1], nullStatus, annotationName);
            } 
            break;
          case 2048:
            nullStatus = flowInfo.nullStatus(local);
            if (nullStatus != 4) {
              FakedTrackingVariable closeTracker = local.closeTracker;
              if (closeTracker == null || 
                closeTracker.hasDefinitelyNoResource(flowInfo))
                break; 
              if (closeTracker.isClosedInFinallyOfEnclosing(scope))
                break; 
              nullStatus = closeTracker.findMostSpecificStatus(flowInfo, scope, null);
              closeTracker.recordErrorLocation(this.nullReferences[j], nullStatus);
              closeTracker.reportRecordedErrors((Scope)scope, nullStatus, (flowInfo.reachMode() != 0));
              this.nullReferences[j] = null;
            } 
            break;
          case 16:
            checkUnboxing((Scope)scope, (Expression)location, flowInfo);
            break;
        } 
      } 
    } 
    if (updateInitsOnBreak) {
      this.initsOnBreak.addPotentialNullInfoFrom(incomingInfo);
      for (int j = 0; j < this.breakTargetsCount; j++)
        (this.breakTargetContexts[j]).initsOnBreak.addPotentialNullInfoFrom(incomingInfo); 
    } 
  }
  
  public BranchLabel continueLabel() {
    return this.continueLabel;
  }
  
  public String individualToString() {
    StringBuffer buffer = new StringBuffer("Looping flow context");
    buffer.append("[initsOnBreak - ").append(this.initsOnBreak.toString()).append(']');
    buffer.append("[initsOnContinue - ").append(this.initsOnContinue.toString()).append(']');
    buffer.append("[finalAssignments count - ").append(this.assignCount).append(']');
    buffer.append("[nullReferences count - ").append(this.nullCount).append(']');
    return buffer.toString();
  }
  
  public boolean isContinuable() {
    return true;
  }
  
  public boolean isContinuedTo() {
    return (this.initsOnContinue != FlowInfo.DEAD_END);
  }
  
  public void recordBreakTo(FlowContext targetContext) {
    if (targetContext instanceof LabelFlowContext) {
      int current;
      if ((current = this.breakTargetsCount++) == 0) {
        this.breakTargetContexts = new LabelFlowContext[2];
      } else if (current == this.breakTargetContexts.length) {
        System.arraycopy(this.breakTargetContexts, 0, this.breakTargetContexts = new LabelFlowContext[current + 2], 0, current);
      } 
      this.breakTargetContexts[current] = (LabelFlowContext)targetContext;
    } 
  }
  
  public void recordContinueFrom(FlowContext innerFlowContext, FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) == 0) {
      if ((this.initsOnContinue.tagBits & 0x1) == 0) {
        this.initsOnContinue = this.initsOnContinue
          .mergedWith(flowInfo.unconditionalInitsWithoutSideEffect());
      } else {
        this.initsOnContinue = flowInfo.unconditionalCopy();
      } 
      FlowContext inner = innerFlowContext;
      while (inner != this && !(inner instanceof org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext))
        inner = inner.parent; 
      if (inner == this) {
        this.upstreamNullFlowInfo
          .addPotentialNullInfoFrom(
            flowInfo.unconditionalInitsWithoutSideEffect());
      } else {
        int length = 0;
        if (this.innerFlowContexts == null) {
          this.innerFlowContexts = new org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext[5];
          this.innerFlowInfos = new UnconditionalFlowInfo[5];
        } else if (this.innerFlowContextsCount == (
          length = this.innerFlowContexts.length) - 1) {
          System.arraycopy(this.innerFlowContexts, 0, 
              this.innerFlowContexts = new org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext[length + 5], 
              0, length);
          System.arraycopy(this.innerFlowInfos, 0, 
              this.innerFlowInfos = new UnconditionalFlowInfo[length + 5], 
              0, length);
        } 
        this.innerFlowContexts[this.innerFlowContextsCount] = (org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext)inner;
        this.innerFlowInfos[this.innerFlowContextsCount++] = 
          flowInfo.unconditionalInitsWithoutSideEffect();
      } 
    } 
  }
  
  protected boolean recordFinalAssignment(VariableBinding binding, Reference finalAssignment) {
    if (binding instanceof LocalVariableBinding) {
      BlockScope blockScope = ((LocalVariableBinding)binding).declaringScope;
      Scope scope;
      while ((scope = ((Scope)blockScope).parent) != null) {
        if (scope == this.associatedScope)
          return false; 
      } 
    } 
    if (this.assignCount == 0) {
      this.finalAssignments = new Reference[5];
      this.finalVariables = new VariableBinding[5];
    } else {
      if (this.assignCount == this.finalAssignments.length)
        System.arraycopy(
            this.finalAssignments, 
            0, 
            this.finalAssignments = new Reference[this.assignCount * 2], 
            0, 
            this.assignCount); 
      System.arraycopy(
          this.finalVariables, 
          0, 
          this.finalVariables = new VariableBinding[this.assignCount * 2], 
          0, 
          this.assignCount);
    } 
    this.finalAssignments[this.assignCount] = finalAssignment;
    this.finalVariables[this.assignCount++] = binding;
    return true;
  }
  
  protected void recordNullReference(LocalVariableBinding local, ASTNode expression, int checkType, FlowInfo nullInfo) {
    if (this.nullCount == 0) {
      this.nullLocals = new LocalVariableBinding[5];
      this.nullReferences = new ASTNode[5];
      this.nullCheckTypes = new int[5];
      this.nullInfos = new UnconditionalFlowInfo[5];
    } else if (this.nullCount == this.nullLocals.length) {
      System.arraycopy(this.nullLocals, 0, 
          this.nullLocals = new LocalVariableBinding[this.nullCount * 2], 0, this.nullCount);
      System.arraycopy(this.nullReferences, 0, 
          this.nullReferences = new ASTNode[this.nullCount * 2], 0, this.nullCount);
      System.arraycopy(this.nullCheckTypes, 0, 
          this.nullCheckTypes = new int[this.nullCount * 2], 0, this.nullCount);
      System.arraycopy(this.nullInfos, 0, 
          this.nullInfos = new UnconditionalFlowInfo[this.nullCount * 2], 0, this.nullCount);
    } 
    this.nullLocals[this.nullCount] = local;
    this.nullReferences[this.nullCount] = expression;
    this.nullCheckTypes[this.nullCount] = checkType;
    this.nullInfos[this.nullCount++] = (nullInfo != null) ? nullInfo.unconditionalCopy() : null;
  }
  
  public void recordUnboxing(Scope scope, Expression expression, int nullStatus, FlowInfo flowInfo) {
    if (nullStatus == 2) {
      super.recordUnboxing(scope, expression, nullStatus, flowInfo);
    } else {
      recordNullReference(null, (ASTNode)expression, 16, flowInfo);
    } 
  }
  
  public boolean recordExitAgainstResource(BlockScope scope, FlowInfo flowInfo, FakedTrackingVariable trackingVar, ASTNode reference) {
    LocalVariableBinding local = trackingVar.binding;
    if (flowInfo.isDefinitelyNonNull(local))
      return false; 
    if (flowInfo.isDefinitelyNull(local)) {
      scope.problemReporter().unclosedCloseable(trackingVar, reference);
      return true;
    } 
    if (flowInfo.isPotentiallyNull(local)) {
      scope.problemReporter().potentiallyUnclosedCloseable(trackingVar, reference);
      return true;
    } 
    recordNullReference(trackingVar.binding, reference, 2048, flowInfo);
    return true;
  }
  
  public void recordUsingNullReference(Scope scope, LocalVariableBinding local, ASTNode location, int checkType, FlowInfo flowInfo) {
    Expression reference;
    if ((flowInfo.tagBits & 0x3) != 0 || 
      flowInfo.isDefinitelyUnknown(local))
      return; 
    checkType |= this.tagBits & 0x1000;
    int checkTypeWithoutHideNullWarning = checkType & 0xFFFF0FFF;
    switch (checkTypeWithoutHideNullWarning) {
      case 256:
      case 512:
        reference = (Expression)location;
        if (flowInfo.isDefinitelyNonNull(local)) {
          if (checkTypeWithoutHideNullWarning == 512) {
            if ((this.tagBits & 0x1000) == 0)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, (ASTNode)reference); 
            flowInfo.initsWhenFalse().setReachMode(2);
          } else {
            scope.problemReporter().localVariableNonNullComparedToNull(local, (ASTNode)reference);
            flowInfo.initsWhenTrue().setReachMode(2);
          } 
        } else if (flowInfo.isDefinitelyNull(local)) {
          if (checkTypeWithoutHideNullWarning == 256) {
            if ((this.tagBits & 0x1000) == 0)
              scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)reference); 
            flowInfo.initsWhenFalse().setReachMode(2);
          } else {
            scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)reference);
            flowInfo.initsWhenTrue().setReachMode(2);
          } 
        } else if (this.upstreamNullFlowInfo.isDefinitelyNonNull(local) && !flowInfo.isPotentiallyNull(local) && !flowInfo.isPotentiallyUnknown(local)) {
          recordNullReference(local, (ASTNode)reference, checkType, flowInfo);
          flowInfo.markAsDefinitelyNonNull(local);
        } else {
          if (flowInfo.cannotBeDefinitelyNullOrNonNull(local))
            return; 
          if (flowInfo.isPotentiallyNonNull(local)) {
            recordNullReference(local, (ASTNode)reference, 0x2 | checkType & 0xFFFFFF00, flowInfo);
          } else if (flowInfo.isPotentiallyNull(local)) {
            recordNullReference(local, (ASTNode)reference, 0x1 | checkType & 0xFFFFFF00, flowInfo);
          } else {
            recordNullReference(local, (ASTNode)reference, checkType, flowInfo);
          } 
        } 
        return;
      case 257:
      case 513:
      case 769:
      case 1025:
        reference = (Expression)location;
        if (flowInfo.isPotentiallyNonNull(local) || 
          flowInfo.isPotentiallyUnknown(local) || 
          flowInfo.isProtectedNonNull(local))
          return; 
        if (flowInfo.isDefinitelyNull(local)) {
          switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
            case 256:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                return;
              } 
              if ((this.tagBits & 0x1000) == 0)
                scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)reference); 
              flowInfo.initsWhenFalse().setReachMode(2);
              return;
            case 512:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                return;
              } 
              scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)reference);
              flowInfo.initsWhenTrue().setReachMode(2);
              return;
            case 768:
              scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)reference);
              return;
            case 1024:
              scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)reference);
              return;
          } 
        } else if (flowInfo.isPotentiallyNull(local)) {
          switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
            case 256:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                return;
              } 
              break;
            case 512:
              if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                return;
              } 
              break;
          } 
        } 
        recordNullReference(local, (ASTNode)reference, checkType, flowInfo);
        return;
      case 3:
        if (flowInfo.isDefinitelyNonNull(local))
          return; 
        if (flowInfo.isDefinitelyNull(local)) {
          scope.problemReporter().localVariableNullReference(local, location);
          return;
        } 
        if (flowInfo.isPotentiallyNull(local)) {
          scope.problemReporter().localVariablePotentialNullReference(local, location);
          return;
        } 
        recordNullReference(local, location, checkType, flowInfo);
        return;
    } 
  }
  
  void removeFinalAssignmentIfAny(Reference reference) {
    for (int i = 0; i < this.assignCount; i++) {
      if (this.finalAssignments[i] == reference) {
        this.finalAssignments[i] = null;
        this.finalVariables[i] = null;
        return;
      } 
    } 
  }
  
  public void simulateThrowAfterLoopBack(FlowInfo flowInfo) {
    if (this.escapingExceptionCatchSites != null) {
      for (int i = 0, exceptionCount = this.escapingExceptionCatchSites.size(); i < exceptionCount; i++)
        ((EscapingExceptionCatchSite)this.escapingExceptionCatchSites.get(i)).simulateThrowAfterLoopBack(flowInfo); 
      this.escapingExceptionCatchSites = null;
    } 
  }
  
  public void recordCatchContextOfEscapingException(ExceptionHandlingFlowContext catchingContext, ReferenceBinding caughtException, FlowInfo exceptionInfo) {
    if (this.escapingExceptionCatchSites == null)
      this.escapingExceptionCatchSites = new ArrayList(5); 
    this.escapingExceptionCatchSites.add(new EscapingExceptionCatchSite(catchingContext, caughtException, exceptionInfo));
  }
  
  public boolean hasEscapingExceptions() {
    return (this.escapingExceptionCatchSites != null);
  }
  
  protected boolean internalRecordNullityMismatch(Expression expression, TypeBinding providedType, FlowInfo flowInfo, int nullStatus, TypeBinding expectedType, int checkType) {
    recordProvidedExpectedTypes(providedType, expectedType, this.nullCount);
    recordNullReference(expression.localVariableBinding(), (ASTNode)expression, checkType, flowInfo);
    return true;
  }
}
